import contextvars
import json
from dataclasses import dataclass
from typing import Optional

from cadence.cadence_types import WorkflowExecution, RecordActivityTaskHeartbeatRequest, ActivityType, \
    PollForActivityTaskResponse, RespondActivityTaskFailedRequest, RespondActivityTaskCompletedRequest
from cadence.exception_handling import serialize_exception
from cadence.exceptions import ActivityCancelledException
from cadence.workflowservice import WorkflowService

current_activity_context = contextvars.ContextVar("current_activity_context")


@dataclass
class ActivityTask:
    @classmethod
    def from_poll_for_activity_task_response(cls, task: PollForActivityTaskResponse) -> 'ActivityTask':
        activity_task: 'ActivityTask' = cls()
        activity_task.task_token = task.task_token
        activity_task.workflow_execution = task.workflow_execution
        activity_task.activity_id = task.activity_id
        activity_task.activity_type = task.activity_type
        activity_task.scheduled_timestamp = task.scheduled_timestamp
        activity_task.schedule_to_close_timeout_seconds = task.schedule_to_close_timeout_seconds
        activity_task.start_to_close_timeout_seconds = task.start_to_close_timeout_seconds
        activity_task.heartbeat_timeout_seconds = task.heartbeat_timeout_seconds
        activity_task.attempt = task.attempt
        activity_task.heartbeat_details = task.heartbeat_details
        activity_task.workflow_domain = task.workflow_domain
        return activity_task

    task_token: bytes = None
    workflow_execution: WorkflowExecution = None
    activity_id: str = None
    activity_type: ActivityType = None
    scheduled_timestamp: int = None
    schedule_to_close_timeout_seconds: int = None
    start_to_close_timeout_seconds: int = None
    heartbeat_timeout_seconds: int = None
    attempt: int = None
    heartbeat_details: bytes = None
    workflow_domain: str = None


def heartbeat(service: WorkflowService, task_token: bytes, details: object):
    request = RecordActivityTaskHeartbeatRequest()
    request.details = json.dumps(details).encode("utf-8")
    request.identity = WorkflowService.get_identity()
    request.task_token = task_token
    response, error = service.record_activity_task_heartbeat(request)
    if error:
        raise error
    if response.cancel_requested:
        raise ActivityCancelledException()


def get_heartbeat_details(heartbeat_details) -> object:
    if not heartbeat_details:
        return None
    json_text = heartbeat_details.decode("utf-8")
    return json.loads(json_text)


class ActivityContext:
    service: WorkflowService = None
    activity_task: ActivityTask = None
    domain: str = None
    do_not_complete: bool = False

    @staticmethod
    def get() -> 'ActivityContext':
        return current_activity_context.get()

    @staticmethod
    def set(context: 'ActivityContext'):
        current_activity_context.set(context)

    def heartbeat(self, details: object):
        heartbeat(self.service, self.activity_task.task_token, details)

    def get_heartbeat_details(self) -> object:
        return get_heartbeat_details(self.activity_task.heartbeat_details)

    def do_not_complete_on_return(self):
        self.do_not_complete = True


class Activity:

    @staticmethod
    def get_task_token() -> bytes:
        return ActivityContext.get().activity_task.task_token

    @staticmethod
    def get_workflow_execution() -> WorkflowExecution:
        return ActivityContext.get().activity_task.workflow_execution

    @staticmethod
    def get_domain() -> str:
        return ActivityContext.get().domain

    @staticmethod
    def get_heartbeat_details() -> object:
        return ActivityContext.get().get_heartbeat_details()

    @staticmethod
    def heartbeat(details: object):
        ActivityContext.get().heartbeat(details)

    @staticmethod
    def get_activity_task() -> ActivityTask:
        return ActivityContext.get().activity_task

    @staticmethod
    def do_not_complete_on_return():
        return ActivityContext.get().do_not_complete_on_return()


@dataclass
class ActivityCompletionClient:
    service: WorkflowService

    def heartbeat(self, task_token: bytes, details: object):
        heartbeat(self.service, task_token, details)

    def complete(self, task_token: bytes, return_value: object):
        error = complete(self.service, task_token, return_value)
        if error:
            raise error

    def complete_exceptionally(self, task_token: bytes, ex: Exception):
        error = complete_exceptionally(self.service, task_token, ex)
        if error:
            raise error


def complete_exceptionally(service, task_token, ex: Exception) -> Optional[Exception]:
    respond: RespondActivityTaskFailedRequest = RespondActivityTaskFailedRequest()
    respond.task_token = task_token
    respond.identity = WorkflowService.get_identity()
    respond.reason = "ActivityFailureException"
    respond.details = serialize_exception(ex)
    _, error = service.respond_activity_task_failed(respond)
    return error


def complete(service, task_token, return_value: object) -> Optional[Exception]:
    respond = RespondActivityTaskCompletedRequest()
    respond.task_token = task_token
    respond.result = json.dumps(return_value)
    respond.identity = WorkflowService.get_identity()
    _, error = service.respond_activity_task_completed(respond)
    return error



import inspect
from typing import Callable


def get_replay_aware_interceptor(fn: Callable):
    def interceptor(*args, **kwargs):
        from cadence.decision_loop import ITask
        task: ITask = ITask.current()
        if not task.decider.decision_context.is_replaying():
            return fn(*args, **kwargs)

    return interceptor


def make_replay_aware(target: object):
    # TODO: Consider using metaclasses instead
    if hasattr(target, "_cadence_python_intercepted"):
        return target
    for name, fn in inspect.getmembers(target):
        if inspect.ismethod(fn):
            setattr(target, name, get_replay_aware_interceptor(fn))
    target._cadence_python_intercepted = True
    return target



from dataclasses import dataclass

from cadence.cadence_types import TimeoutType, ActivityType, WorkflowExecution, WorkflowExecutionCloseStatus
from cadence.exception_handling import deserialize_exception


class IllegalStateException(BaseException):
    pass


class IllegalArgumentException(BaseException):
    pass


class WorkflowTypeNotFound(Exception):
    pass


class NonDeterministicWorkflowException(BaseException):
    pass


class ActivityTaskFailedException(Exception):

    def __init__(self, reason: str, cause: Exception) -> None:
        super().__init__(reason)
        self.reason = reason
        self.cause = cause


class ActivityTaskTimeoutException(Exception):

    def __init__(self, event_id: int, timeout_type: TimeoutType, details: bytes, *args: object) -> None:
        super().__init__(*args)
        self.details = details
        self.timeout_type = timeout_type
        self.event_id = event_id


class SignalNotFound(Exception):
    pass


class QueryNotFound(Exception):
    pass


class QueryDidNotComplete(Exception):
    pass

class CancellationException(Exception):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cause = None

    def init_cause(self, cause):
        self.cause = cause


class ActivityCancelledException(Exception):
    pass


class WorkflowOperationException(Exception):
    def __init__(self, event_id: int):
        self.event_id = event_id


class ActivityException(WorkflowOperationException):
    def __init__(self, event_id: int, activity_type: str, activity_id: str):
        super().__init__(event_id=event_id)
        self.activity_type = activity_type
        self.activity_id = activity_id


class ActivityFailureException(ActivityException):
    def __init__(self, event_id: int, activity_type: str, activity_id: str, cause: str):
        super().__init__(event_id, activity_type, activity_id)
        self.cause: str = cause
        self.attempt: int = None
        self.backoff: int = 0

    def set_cause(self):
        if self.cause:
            cause_ex = deserialize_exception(self.cause)
            self.__cause__ = cause_ex

    def get_cause(self):
        if self.cause:
            return deserialize_exception(self.cause)
        else:
            return None


@dataclass
class WorkflowException(Exception):
    workflow_type: str = None
    execution: WorkflowExecution = None

    def __str__(self):
        return f'{type(self).__name__}: WorkflowType="{self.workflow_type}", ' \
               f'WorkflowID="{self.execution.workflow_id}", RunID="{self.execution.run_id} '


@dataclass
class WorkflowFailureException(WorkflowException):
    decision_task_completed_event_id: int = None


@dataclass
class QueryFailureException(Exception):
    query_type: str = None
    execution: WorkflowExecution = None

    def __str__(self):
        return f'{type(self).__name__}: QueryType="{self.query_type}", ' \
               f'WorkflowID="{self.execution.workflow_id}", RunID="{self.execution.run_id} '



class QueryRejectedException(Exception):
    close_status: WorkflowExecutionCloseStatus

    def __init__(self, close_status: WorkflowExecutionCloseStatus):
        self.close_status = close_status



from __future__ import annotations

from dataclasses import dataclass, field
from dataclasses_json import dataclass_json, LetterCase

from typing import Dict, Optional

from cadence.cadence_types import Header, EventType, MarkerRecordedEventAttributes, HistoryEvent
from cadence.decision_loop import DecisionContext

MUTABLE_MARKER_HEADER_KEY = "MutableMarkerHeader"


class MarkerInterface:
    @staticmethod
    def from_event_attributes(attributes: MarkerRecordedEventAttributes) -> MarkerInterface:
        if attributes.header and attributes.header.fields and MUTABLE_MARKER_HEADER_KEY in attributes.header.fields:
            buffer = attributes.header.fields.get(MUTABLE_MARKER_HEADER_KEY)
            header = MarkerHeader.from_json(str(buffer, "utf-8"))
            return MarkerData(header=header, data=attributes.details)
        return PlainMarkerData.from_json(str(attributes.details, "utf-8"))

    def get_id(self) -> str:
        raise NotImplementedError()

    def get_access_count(self) -> int:
        raise NotImplementedError()

    def get_data(self) -> bytes:
        raise NotImplementedError()


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class MarkerHeader:
    id: str = None
    event_id: int = None
    access_count: int = 0


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class MarkerData(MarkerInterface):
    header: MarkerHeader = None
    data: bytes = None

    @staticmethod
    def create(id: str, event_id: int, data: bytes, access_count: int) -> MarkerData:
        header = MarkerHeader(id=id, event_id=event_id, access_count=access_count)
        return MarkerData(header=header, data=data)

    def get_header(self) -> Header:
        header_bytes = self.header.to_json().encode("utf-8")
        header = Header()
        header.fields[MUTABLE_MARKER_HEADER_KEY] = header_bytes
        return header

    def get_access_count(self) -> int:
        return self.header.access_count

    def get_data(self) -> bytes:
        return self.data

    def get_id(self) -> str:
        return self.header.id


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class MarkerResult:
    data: bytes = None
    access_count: int = 0
    replayed: bool = False


@dataclass
class MarkerHandler:
    decision_context: DecisionContext
    marker_name: str
    mutable_marker_results: Dict[str, MarkerResult] = field(default_factory=dict)

    def record_mutable_marker(self, id: str, event_id: int, data: bytes, access_count: int):
        marker = MarkerData.create(id=id, event_id=event_id, data=data, access_count=access_count)
        if id in self.mutable_marker_results:
            self.mutable_marker_results[id].replayed = True
        else:
            self.mutable_marker_results[id] = MarkerResult(data=data)
        self.decision_context.record_marker(self.marker_name, marker.get_header(), data)

    # Sets data without creating a decision - used when DEFAULT_VERSION is the implicit current version
    def set_data(self, id, data: bytes):
        self.mutable_marker_results[id] = MarkerResult(data=data)

    def mark_replayed(self, id):
        self.mutable_marker_results[id].replayed = True

    def handle(self, id: str, func) -> Optional[bytes]:
        event_id = self.decision_context.decider.next_decision_event_id
        result: MarkerResult = self.mutable_marker_results.get(id)
        if result or self.decision_context.is_replaying():
            if result:
                if self.decision_context.is_replaying() and not result.replayed:
                    # Need to insert marker to ensure that event_id is incremented
                    self.record_mutable_marker(id, event_id, result.data, 0)
                return result.data
            else:
                return None
        else:
            to_store = func()
            if to_store:
                data = to_store
                self.record_mutable_marker(id, event_id, data, 0)
                return to_store
            else:
                # TODO: Should this ever happen? - at least for version it will never happen
                pass

    # This method is currently not being used - after adopting the version logic from the
    # Golang client
    def get_marker_data_from_history(self, event_id: int, marker_id: str, expected_access_count: int) -> \
            Optional[bytes]:
        event: HistoryEvent = self.decision_context.decider.get_optional_decision_event(event_id)
        if not event or event.event_type != EventType.MarkerRecorded:
            return None

        attributes: MarkerRecordedEventAttributes = event.marker_recorded_event_attributes
        name = attributes.marker_name
        if self.marker_name != name:
            return None

        marker_data = MarkerInterface.from_event_attributes(attributes)
        if marker_id != marker_data.get_id() or marker_data.get_access_count() > expected_access_count:
            return None

        return marker_data.get_data()


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class PlainMarkerData(MarkerInterface):
    id: str = None
    event_id: int = None
    data: bytes = None
    access_count: int = 0

    def get_access_count(self):
        return self.access_count

    def get_data(self):
        return self.data

    def get_id(self) -> str:
        return self.id




import json
import logging
import traceback

import tblib

THIS_SOURCE = "cadence-python"

logger = logging.getLogger(__name__)


class ExternalException(Exception):
    def __init__(self, details):
        super().__init__(details)

    @property
    def details(self):
        return self.args[0]


def exception_class_fqn(o):
    # Copied from: https://stackoverflow.com/a/2020083
    module = o.__class__.__module__
    if module is None or module == str.__class__.__module__:
        return o.__class__.__name__  # Avoid reporting __builtin__
    else:
        return module + '.' + o.__class__.__name__


def import_class_from_string(path):
    # Taken from: https://stackoverflow.com/a/30042585
    from importlib import import_module
    module_path, _, class_name = path.rpartition('.')
    mod = import_module(module_path)
    klass = getattr(mod, class_name)
    return klass


def serialize_exception(ex: Exception):
    exception_cls_name: str = exception_class_fqn(ex)
    tb = "".join(traceback.format_exception(type(ex), ex, ex.__traceback__))
    details = json.dumps({
        "class": exception_cls_name,
        "args": ex.args,
        "traceback": tb,
        "source": THIS_SOURCE
    })
    return details


def deserialize_exception(details) -> Exception:
    """
    TODO: Support built-in types like Exception
    """
    exception: Exception = None
    details_dict = json.loads(details)
    source = details_dict.get("source")
    exception_cls_name: str = details_dict.get("class")

    if source == THIS_SOURCE and exception_cls_name:
        try:
            klass = import_class_from_string(exception_cls_name)
            exception = klass(*details_dict["args"])
            t = tblib.Traceback.from_string(details_dict["traceback"])
            exception.with_traceback(t.as_traceback())
        except Exception as e:
            exception = None
            logger.error("Failed to deserialize exception (details=%s) cause=%r", details_dict, e)

    if not exception:
        return ExternalException(details_dict)
    else:
        return exception



import datetime
import logging
import json

from cadence.activity import ActivityContext, ActivityTask, complete_exceptionally, complete
from cadence.cadence_types import PollForActivityTaskRequest, TaskListMetadata, TaskList, PollForActivityTaskResponse
from cadence.conversions import json_to_args
from cadence.workflowservice import WorkflowService
from cadence.worker import Worker, StopRequestedException

logger = logging.getLogger(__name__)


def activity_task_loop(worker: Worker):
    service: WorkflowService = WorkflowService.create(worker.host, worker.port, timeout=worker.get_timeout())
    worker.manage_service(service)
    logger.info(f"Activity task worker started: {WorkflowService.get_identity()}")
    try:
        while True:
            if worker.is_stop_requested():
                return
            try:
                service.set_next_timeout_cb(worker.raise_if_stop_requested)

                polling_start = datetime.datetime.now()
                polling_request = PollForActivityTaskRequest()
                polling_request.task_list_metadata = TaskListMetadata()
                polling_request.task_list_metadata.max_tasks_per_second = 200000
                polling_request.domain = worker.domain
                polling_request.identity = WorkflowService.get_identity()
                polling_request.task_list = TaskList()
                polling_request.task_list.name = worker.task_list
                task: PollForActivityTaskResponse
                task, err = service.poll_for_activity_task(polling_request)
                polling_end = datetime.datetime.now()
                logger.debug("PollForActivityTask: %dms", (polling_end - polling_start).total_seconds() * 1000)
            except StopRequestedException:
                return
            except Exception as ex:
                logger.error("PollForActivityTask error: %s", ex)
                continue
            if err:
                logger.error("PollForActivityTask failed: %s", err)
                continue
            task_token = task.task_token
            if not task_token:
                logger.debug("PollForActivityTask has no task_token (expected): %s", task)
                continue

            args = json_to_args(task.input)
            logger.info(f"Request for activity: {task.activity_type.name}")
            fn = worker.activities.get(task.activity_type.name)
            if not fn:
                logger.error("Activity type not found: " + task.activity_type.name)
                continue

            process_start = datetime.datetime.now()
            activity_context = ActivityContext()
            activity_context.service = service
            activity_context.activity_task = ActivityTask.from_poll_for_activity_task_response(task)
            activity_context.domain = worker.domain
            try:
                ActivityContext.set(activity_context)
                return_value = fn(*args)
                if activity_context.do_not_complete:
                    logger.info(f"Not completing activity {task.activity_type.name}({str(args)[1:-1]})")
                    continue
                error = complete(service, task_token, return_value)
                if error:
                    logger.error("Error invoking RespondActivityTaskCompleted: %s", error)
                logger.info(f"Activity {task.activity_type.name}({str(args)[1:-1]}) returned {json.dumps(return_value)}")
            except Exception as ex:
                logger.error(f"Activity {task.activity_type.name} failed: {type(ex).__name__}({ex})", exc_info=1)
                error = complete_exceptionally(service, task_token, ex)
                if error:
                    logger.error("Error invoking RespondActivityTaskFailed: %s", error)
            finally:
                ActivityContext.set(None)
                process_end = datetime.datetime.now()
                logger.info("Process ActivityTask: %dms", (process_end - process_start).total_seconds() * 1000)
    finally:
        try:
            service.close()
        except:
            logger.warning("service.close() failed", exc_info=1)
        worker.notify_thread_stopped()



from select import select
from socket import socket
from typing import IO, Callable


class IOWrapper:
    def __init__(self, io_stream: IO, socket_: socket = None):
        self.io_stream = io_stream
        self.socket = socket_
        self.next_timeout_cb = None

    def set_next_timeout_cb(self, cb: Callable):
        self.next_timeout_cb = cb

    def read_or_eof(self, size, field):
        if self.next_timeout_cb and self.socket:
            timeout = self.socket.gettimeout()
            self.socket.setblocking(False)
            while True:
                ready_to_read, _, _ = select([self.socket], [], [], 1)
                if ready_to_read:
                    break
                else:
                    self.next_timeout_cb()
            self.next_timeout_cb = None
            self.socket.setblocking(True)
            self.socket.settimeout(timeout)
        buf: bytes = self.io_stream.read(size)
        if len(buf) != size:
            raise EOFError(field)
        return buf

    def read_short(self, field: str) -> int:
        return int.from_bytes(self.read_or_eof(2, field), byteorder='big', signed=False)

    def read_long(self, field: str) -> int:
        return int.from_bytes(self.read_or_eof(4, field), byteorder='big', signed=False)

    def read_byte(self, field: str) -> int:
        return int.from_bytes(self.read_or_eof(1, field), byteorder='big', signed=False)

    def read_bytes(self, n: int, field: str) -> bytes:
        buf: bytes = self.read_or_eof(n, field)
        return buf

    def read_string(self, n: int, field: str) -> str:
        buf: bytes = self.read_or_eof(n, field)
        return str(buf, "utf-8")

    def write_short(self, v: int):
        self.io_stream.write(v.to_bytes(2, byteorder='big', signed=False))

    def write_long(self, v: int):
        self.io_stream.write(v.to_bytes(4, byteorder='big', signed=False))

    def write_byte(self, v: int):
        self.io_stream.write(v.to_bytes(1, byteorder='big', signed=False))

    def write_bytes(self, b: bytes):
        self.io_stream.write(b)

    def write_string(self, s: str):
        self.io_stream.write(bytes(s, "utf-8"))

    def flush(self):
        self.io_stream.flush()

    def close(self):
        self.io_stream.close()
